@@ -8,6 +8,7 @@ use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION
             $def_timeout $def_proto $def_factor $def_family
             $max_datasize $pingstring $hires $source_verify $syn_forking);
 use Fcntl qw( F_GETFL F_SETFL O_NONBLOCK );
+use Socket 2.007;
 use Socket qw( SOCK_DGRAM SOCK_STREAM SOCK_RAW AF_INET PF_INET IPPROTO_TCP
 	       SOL_SOCKET SO_ERROR SO_BROADCAST
                IPPROTO_IP IP_TOS IP_TTL
@@ -21,7 +22,7 @@ use Time::HiRes;
 @ISA = qw(Exporter);
 @EXPORT = qw(pingecho);
 @EXPORT_OK = qw(wakeonlan);
-$VERSION = "2.73";
+$VERSION = "2.76";
 
 # Globals
 
@@ -46,7 +47,7 @@ my $NIx_NOSERV = eval { Socket::NIx_NOSERV() } || 2;
 #my $IPV6_HOPLIMIT  = eval { Socket::IPV6_HOPLIMIT() };  # ping6 -h 0-255
 my $qr_family = qr/^(?:(?:(:?ip)?v?(?:4|6))|${\AF_INET}|$AF_INET6)$/;
 my $qr_family4 = qr/^(?:(?:(:?ip)?v?4)|${\AF_INET})$/;
-my $Socket_VERSION = eval { $Socket::VERSION };
+my $Socket_VERSION = eval $Socket::VERSION;
 
 if ($^O =~ /Win32/i) {
   # Hack to avoid this Win32 spewage:
@@ -226,13 +227,18 @@ sub new
   }
   elsif ($self->{proto} eq "icmp")
   {
-    croak("icmp ping requires root privilege") if !_isroot();
+    croak("icmp ping requires root privilege") if !_isroot() and $^O ne "linux";
     $self->{proto_num} = eval { (getprotobyname('icmp'))[2] } ||
       croak("Can't get icmp protocol by name");
     $self->{pid} = $$ & 0xffff;           # Save lower 16 bits of pid
     $self->{fh} = FileHandle->new();
-    socket($self->{fh}, PF_INET, SOCK_RAW, $self->{proto_num}) ||
-      croak("icmp socket error - $!");
+    if ($^O eq "linux" and !_isroot()) {
+      socket($self->{fh}, PF_INET, SOCK_DGRAM, $self->{proto_num}) ||
+          croak("icmp socket error - $!");
+    } else {
+      socket($self->{fh}, PF_INET, SOCK_RAW, $self->{proto_num}) ||
+          croak("icmp socket error - $!");
+    }
     $self->_setopts();
     if ($self->{'ttl'}) {
       setsockopt($self->{fh}, IPPROTO_IP, IP_TTL, pack("I*", $self->{'ttl'}))
@@ -249,8 +255,13 @@ sub new
       croak("Can't get ipv6-icmp protocol by name"); # 58
     $self->{pid} = $$ & 0xffff;           # Save lower 16 bits of pid
     $self->{fh} = FileHandle->new();
-    socket($self->{fh}, $AF_INET6, SOCK_RAW, $self->{proto_num}) ||
-      croak("icmp socket error - $!");
+    if ($^O eq 'linux' and !_isroot()) {
+      socket($self->{fh}, $AF_INET6, SOCK_DGRAM, $self->{proto_num}) ||
+          croak("icmp socket error - $!");
+    } else {
+      socket($self->{fh}, $AF_INET6, SOCK_RAW, $self->{proto_num}) ||
+          croak("icmp socket error - $!");
+    }
     $self->_setopts();
     if ($self->{'gateway'}) {
       my $g = $self->{gateway};
@@ -644,10 +655,11 @@ sub ping_external {
 # h2ph "asm/socket.h"
 # require "asm/socket.ph";
 use constant SO_BINDTODEVICE  => 25;
-use constant ICMP_ECHOREPLY   => 0; # ICMP packet types
+use constant ICMP_ECHOREPLY   => 0;   # ICMP packet types
 use constant ICMPv6_ECHOREPLY => 129; # ICMP packet types
-use constant ICMP_UNREACHABLE => 3; # ICMP packet types
+use constant ICMP_UNREACHABLE => 3;   # ICMP packet types
 use constant ICMPv6_UNREACHABLE => 1; # ICMP packet types
+use constant ICMPv6_NI_REPLY => 140;  # ICMP packet types
 use constant ICMP_ECHO        => 8;
 use constant ICMPv6_ECHO      => 128;
 use constant ICMP_TIME_EXCEEDED => 11; # ICMP packet types
@@ -713,8 +725,13 @@ sub ping_icmp
   $timeout = $self->{timeout} if !defined $timeout and $self->{timeout};
   $timestamp_msg = $self->{message_type} && $self->{message_type} eq 'timestamp' ? 1 : 0;
 
-  socket($self->{fh}, $ip->{family}, SOCK_RAW, $self->{proto_num}) ||
-    croak("icmp socket error - $!");
+  if ($^O eq 'linux' and !_isroot()) {
+      socket($self->{fh}, $ip->{family}, SOCK_DGRAM, $self->{proto_num}) ||
+          croak("icmp socket error - $!");
+  } else {
+      socket($self->{fh}, $ip->{family}, SOCK_RAW, $self->{proto_num}) ||
+          croak("icmp socket error - $!");
+  }
 
   if (defined $self->{local_addr} &&
       !CORE::bind($self->{fh}, _pack_sockaddr_in(0, $self->{local_addr}))) {
@@ -781,31 +798,25 @@ sub ping_icmp
       $from_saddr = recv($self->{fh}, $recv_msg, 1500, ICMP_FLAGS);
       $recv_msg_len = length($recv_msg) - length($self->{data});
       ($from_port, $from_ip) = _unpack_sockaddr_in($from_saddr, $ip->{family});
-      ($from_type, $from_subcode) = unpack("C2", substr($recv_msg, 20, 2));
+      # ICMP echo includes the header and ICMPv6 doesn't.
+      # IPv4 length($recv_msg) is 28 (20 header + 8 payload)
+      # while IPv6 length is only 8 (sans header).
+      my $off = ($ip->{family} == AF_INET) ? 20 : 0; # payload offset
+      ($from_type, $from_subcode) = unpack("C2", substr($recv_msg, $off, 2));
       if ($from_type == ICMP_TIMESTAMP_REPLY) {
-        ($from_pid, $from_seq) = unpack("n3", substr($recv_msg, 24, 4))
-          if length $recv_msg >= 28;
-      } elsif ($from_type == ICMP_ECHOREPLY) {
+        ($from_pid, $from_seq) = unpack("n3", substr($recv_msg, $off + 4, 4))
+          if length $recv_msg >= $off + 8;
+      } elsif ($from_type == ICMP_ECHOREPLY || $from_type == ICMPv6_ECHOREPLY) {
         #warn "ICMP_ECHOREPLY: ", $ip->{family}, " ",$recv_msg, ":", length($recv_msg);
-        ($from_pid, $from_seq) = unpack("n2", substr($recv_msg, 24, 4))
-          if ($ip->{family} == AF_INET && $recv_msg_len == 28);
+        ($from_pid, $from_seq) = unpack("n2", substr($recv_msg, $off + 4, 4))
+          if $recv_msg_len == $off + 8;
+      } elsif ($from_type == ICMPv6_NI_REPLY) {
         ($from_pid, $from_seq) = unpack("n2", substr($recv_msg, 4, 4))
-          if ($ip->{family} == $AF_INET6 && $recv_msg_len == 8);
-      } elsif ($from_type == ICMPv6_ECHOREPLY) {
-        #($from_pid, $from_seq) = unpack("n3", substr($recv_msg, 24, 4))
-        #  if length $recv_msg >= 28;
-        #($from_pid, $from_seq) = unpack("n2", substr($recv_msg, 24, 4))
-        #  if ($ip->{family} == AF_INET && length $recv_msg == 28);
-        #warn "ICMPv6_ECHOREPLY: ", $ip->{family}, " ",$recv_msg, ":", length($recv_msg);
-        ($from_pid, $from_seq) = unpack("n2", substr($recv_msg, 4, 4))
-          if ($ip->{family} == $AF_INET6 && $recv_msg_len == 8);
-      #} elsif ($from_type == ICMPv6_NI_REPLY) {
-      #  ($from_pid, $from_seq) = unpack("n2", substr($recv_msg, 4, 4))
-      #    if ($ip->{family} == $AF_INET6 && length $recv_msg == 8);
+          if ($ip->{family} == $AF_INET6 && length $recv_msg == 8);
       } else {
         #warn "ICMP: ", $from_type, " ",$ip->{family}, " ",$recv_msg, ":", length($recv_msg);
-        ($from_pid, $from_seq) = unpack("n2", substr($recv_msg, 52, 4))
-          if length $recv_msg >= 56;
+        ($from_pid, $from_seq) = unpack("n2", substr($recv_msg, $off + 32, 4))
+          if length $recv_msg >= $off + 36;
       }
       $self->{from_ip} = $from_ip;
       $self->{from_type} = $from_type;
@@ -1085,8 +1096,7 @@ sub tcp_connect
 
 sub DESTROY {
   my $self = shift;
-  if ($self->{'proto'} eq 'tcp' &&
-      $self->{'tcp_chld'}) {
+  if ($self->{'proto'} && ($self->{'proto'} eq 'tcp') && $self->{'tcp_chld'}) {
     # Put that choking client out of its misery
     kill "KILL", $self->{'tcp_chld'};
     # Clean off the zombie
@@ -2008,13 +2018,13 @@ Net::Ping - check a remote host for reachability
 
     use Net::Ping;
 
-    $p = Net::Ping->new();
+    my $p = Net::Ping->new();
     print "$host is alive.\n" if $p->ping($host);
     $p->close();
 
-    $p = Net::Ping->new("icmp");
+    my $p = Net::Ping->new("icmp");
     $p->bind($my_addr); # Specify source interface of pings
-    foreach $host (@host_array)
+    foreach my $host (@host_array)
     {
         print "$host is ";
         print "NOT " unless $p->ping($host, 2);
@@ -2023,7 +2033,11 @@ Net::Ping - check a remote host for reachability
     }
     $p->close();
 
-    $p = Net::Ping->new("tcp", 2);
+    my $p = Net::Ping->new("icmpv6");
+    my $ip = "[fd00:dead:beef::4e]";
+    print "$ip is alive.\n" if $p->ping($ip);
+
+    my $p = Net::Ping->new("tcp", 2);
     # Try connecting to the www port instead of the echo port
     $p->port_number(scalar(getservbyname("http", "tcp")));
     while ($stop_time > time())
@@ -2035,19 +2049,19 @@ Net::Ping - check a remote host for reachability
     undef($p);
 
     # Like tcp protocol, but with many hosts
-    $p = Net::Ping->new("syn");
+    my $p = Net::Ping->new("syn");
     $p->port_number(getservbyname("http", "tcp"));
-    foreach $host (@host_array) {
+    foreach my $host (@host_array) {
       $p->ping($host);
     }
-    while (($host,$rtt,$ip) = $p->ack) {
+    while (my ($host, $rtt, $ip) = $p->ack) {
       print "HOST: $host [$ip] ACKed in $rtt seconds.\n";
     }
 
     # High precision syntax (requires Time::HiRes)
-    $p = Net::Ping->new();
+    my $p = Net::Ping->new();
     $p->hires();
-    ($ret, $duration, $ip) = $p->ping($host, 5.5);
+    my ($ret, $duration, $ip) = $p->ping($host, 5.5);
     printf("$host [ip: $ip] is alive (packet return time: %.2f ms)\n",
             1000 * $duration)
       if $ret;
@@ -2367,11 +2381,13 @@ enabled.
 X<ping_icmp>
 
 The L</ping> method used with the icmp protocol.
+Under Linux under a non-root account this uses now SOCK_DGRAM.
 
-=item $p->ping_icmpv6([$host, $timeout, $family]) I<NYI>
+=item $p->ping_icmpv6([$host, $timeout, $family])
 X<ping_icmpv6>
 
 The L</ping> method used with the icmpv6 protocol.
+Under Linux under a non-root account this uses now SOCK_DGRAM.
 
 =item $p->ping_stream([$host, $timeout, $family])
 X<ping_stream>
@@ -2574,7 +2590,7 @@ L<https://github.com/rurban/Net-Ping/issues>
 
 =head1 COPYRIGHT
 
-Copyright (c) 2017-2018, Reini Urban.  All rights reserved.
+Copyright (c) 2017-2020, Reini Urban.  All rights reserved.
 
 Copyright (c) 2016, cPanel Inc.  All rights reserved.
 
